"""Module for interacting with Git."""

import os
from dataclasses import dataclass
import re
from typing import List, Optional
from pagekey_semver.git.effector import CommandGitEffector, GitEffector
from pagekey_semver.git.querier import CommandGitQuerier, GitQuerier
from pagekey_semver.config import SemverConfig
from pagekey_semver.models import Commit, Tag


@dataclass
class LocalGitOptions:
    """Any Git options that will be changes by this package."""

    name: str
    email: str
    remote: str


class GitManagerException(Exception):
    """Exception managing Git."""


class GitManager:
    """Class to handle all communications with Git executable."""

    def __init__(
        self,
        config: SemverConfig,
        querier: GitQuerier = CommandGitQuerier(),
        effector: GitEffector = CommandGitEffector(),
    ):
        """Initialize Git manager."""
        self._config = config
        self._querier = querier
        self._effector = effector

    def get_existing_git_info(self) -> LocalGitOptions:
        """Determine exisitng Git config information.

        This will allow the original configuration to be restored after run,
        which is important to avoid messing up the user's configuration when
        running locally.

        Returns:
            LocalGitOptions representing any Git options that may get changed.

        Raises:
            CalledProcessError if there is an issue calling Git to check these values.
        """
        return LocalGitOptions(
            name=self._querier.get_config_item("user.name"),
            email=self._querier.get_config_item("user.email"),
            remote=self._querier.get_config_item("remote.origin.url"),
        )

    def get_git_tags(self) -> List[str]:
        """Get a list of all Git tags for the current repo.

        Returns:
            List of each tag name as a string.
        """
        return self._querier.get_tag_names()

    def get_commit_messages_since(self, commit_hash: Optional[str]) -> List[Commit]:
        """Return a  list of commit messages since a commit ref.

        Args:
            commit_hash: Reference to a commit to pull messages since, or None.

        Returns:
            Commit messages since `commit_hash` if provided.
            All commit messages if `commit_hash` is None.
        """
        return self._querier.get_commits(commit_hash)

    def apply_tag(self, existing_tags: List[Tag], new_tag: Tag) -> None:
        """Commit, tag, and push.

        Changes Git username, email based on config
        and restores them to what they were previously when done.

        Args:
            existing_tags: List of pre-existing Git tags in repo.
            new_tag: Tag to be added, if it does not already exist.
        """
        if new_tag not in existing_tags:
            print(f"Tagging/pushing new tag: {new_tag}", flush=True)
            original_git_config = self.get_existing_git_info()

            self.set_git_remote()
            self._effector.set_config_item("user.email", self._config.git.email)
            self._effector.set_config_item("user.name", self._config.git.name)
            self._effector.add_all()
            self._effector.create_commit(new_tag.name)
            self._effector.create_tag(new_tag.name)
            self._effector.push("origin", new_tag.name)
            self._effector.push("origin", "HEAD")
            # Restore original Git config
            if len(original_git_config.email) > 0:
                self._effector.set_config_item("user.email", original_git_config.email)
            if len(original_git_config.email) > 0:
                self._effector.set_config_item("user.name", original_git_config.name)
            if len(original_git_config.email) > 0:
                self._effector.set_config_item(
                    "remote.origin.url", original_git_config.remote
                )
        else:
            print(f"Tag {new_tag} already exists - skipping tag/push.", flush=True)

    def set_git_remote(self) -> None:
        """Set push remote based on SEMVER_USER and SEMVER_TOKEN env vars."""
        user = os.getenv("SEMVER_USER", "")
        token = os.getenv("SEMVER_TOKEN", "")
        if len(user) > 0 and len(token) > 0:
            existing_remote = self._querier.get_config_item("remote.origin.url")
            if existing_remote.startswith("git@"):
                # Replace SSH url.
                new_remote = existing_remote.replace(":", "/").replace(
                    "git@", f"https://{user}:{token}@"
                )
            elif existing_remote.startswith("https://"):
                HTTPS_WITH_CREDENTIALS_PATTERN = r"https\://[^:@]+:[^:@]+@"
                if re.search(HTTPS_WITH_CREDENTIALS_PATTERN, existing_remote):
                    # Replace HTTPS with existing auth.
                    new_remote = re.sub(
                        HTTPS_WITH_CREDENTIALS_PATTERN,
                        f"https://{user}:{token}@",
                        existing_remote,
                    )
                else:
                    # Add auth to HTTPS without any credentials.
                    new_remote = existing_remote.replace(
                        "https://", f"https://{user}:{token}@"
                    )
            else:
                raise GitManagerException(
                    f"Unsupported remote URL format: {existing_remote}"
                )
            # Set the new remote.
            self._effector.set_config_item("remote.origin.url", new_remote)
        else:
            print(
                "Warning: SEMVER_USER and/or SEMVER_TOKEN not defined. Consider defining them for push authorization."
            )



"""Module containg data types imported throughout."""

from dataclasses import dataclass
import enum
from pydantic import BaseModel, field_serializer


class ReleaseType(enum.Enum):
    """Logical type of semantic versioning release."""

    NO_RELEASE = "no release"
    PATCH = "patch"
    MINOR = "minor"
    MAJOR = "major"


class GitConfig(BaseModel):
    """Git configuration options to set before committing."""

    name: str
    email: str


class Prefix(BaseModel):
    """Commit prefix that can trigger releases."""

    label: str
    type: ReleaseType

    @field_serializer("type")
    def serialize_type(self, type: ReleaseType, _info):
        """Convert ReleaseType to a string when serializing."""
        return type.value


@dataclass
class Commit:
    """Represents a Git commit."""

    hash: str
    message: str


@dataclass
class Tag:
    """Represents a Semver logical tag."""

    # The formatted tag name as it appears in Git.
    name: str
    # The major part of the version.
    major: int
    # The minor part of the version.
    minor: int
    # The patch part of the version.
    patch: int



"""Module for writing changelogs."""

from __future__ import annotations
import abc
import os
import tempfile
from typing import List, TextIO

from pagekey_semver.util.dynamic_import import dynamic_import
from pagekey_semver.config import SemverConfig
from pagekey_semver.models import Commit, Tag


class ChangelogWriter(abc.ABC):
    """Base class for all ChangelogWriters."""

    @staticmethod
    def from_config(config: SemverConfig) -> ChangelogWriter:
        """Create a ChangelogWriter from a config file.

        Uses the `changelog_writer` string to dynamically import a Changelog class for use.

        Args:
            Config that defines where the ChangelogWriter class is.

        Returns:
            ChangelogWriter instance with imported class, initialized with config.

        Raises:
            ImportError, AttributeError if the import string is not valid.
        """
        changelog_writer_cls = dynamic_import(config.changelog_writer)
        return changelog_writer_cls(config)

    def __init__(self, config: SemverConfig):
        """Initialize the writer.

        Args:
            config: Semver application config.
        """
        self._config = config

    def update_changelog(self, version: Tag, commits: List[Commit]) -> None:
        """Update changelog.

        This is the method called by the CLI to perform the update.

        It contains a few additional methods to simplify the implementation
        of write_changelog, which can be overridden by the user.

        Args:
            version: The new version being added.
            commits: Full list of commits since last release.
        """
        self._create_dirs()
        filtered_commits = self._filter_commits(commits)
        # Write new version info to temp file.
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as temp_file:
            self.write_changelog(temp_file, version, filtered_commits)
            # Write existing changelog to temp file.
            if os.path.exists(self._config.changelog_path):
                with open(self._config.changelog_path, "r") as changelog_file:
                    temp_file.write(changelog_file.read())
            temp_file_name = temp_file.name
        with open(temp_file_name, "r") as temp_file:
            # Copy temp file to changelog, effectively prepending.
            with open(self._config.changelog_path, "w") as changelog_file:
                changelog_file.write(temp_file.read())
        # Delete temp_file
        os.unlink(temp_file_name)

    def _filter_commits(self, commits: List[Commit]) -> List[Commit]:
        """Filter out any commits that do not start with a valid prefix.

        Args:
            commits: Full list of commits since last release.

        Returns:
            List of commits that start with a prefix defined in config.
        """
        result = []
        for commit in commits:
            for prefix in self._config.prefixes:
                if commit.message.startswith(f"{prefix.label}: "):
                    result.append(commit)
        return result

    def _create_dirs(self) -> None:
        """Make directories for Changelog."""
        dirs = os.path.dirname(self._config.changelog_path)
        if len(dirs) > 0:
            os.makedirs(dirs, exist_ok=True)

    @abc.abstractmethod
    def write_changelog(
        self, changelog_file: TextIO, version: Tag, commits: List[Commit]
    ) -> None:
        """Update the changelog file.

        Args:
            version: The new version being added.
            commits: The commits associated with the new version.
        """


class DefaultChangelogWriter(ChangelogWriter):
    """Default concrete implementation of ChangelogWriter.

    Writes a simple CHANGELOG file with a level-2 header for each release
    and a bullet containing message and commit hash for each commit in the release.
    """

    def write_changelog(
        self, changelog_file: TextIO, version: Tag, commits: List[Commit]
    ) -> None:
        """Update the changelog file.

        Args:
            version: The new version being added.
            commits: The commits associated with the new version.
        """
        # Write changelog.
        changelog_file.write(f"## {version.name}\n\n")
        for commit in commits:
            changelog_file.write(f"- {commit.message} ({commit.hash[0:8]})\n")
        changelog_file.write("\n")



"""Module for running commands on the system."""

import abc
import subprocess
from dataclasses import dataclass


@dataclass
class CommandResult:
    """Information returned when executing a command."""

    exit_code: int
    stdout: str
    stderr: str


class CommandRunnerException(Exception):
    """Raised when a command does not complete successfully."""

    def __init__(self, command_result: CommandResult):
        self.command_result = command_result
        super().__init__(
            "Failed to run command: "
            f"returned code {self.command_result.exit_code}\n"
            f"{self.command_result.stdout}\n"
            f"{self.command_result.stderr}"
        )


class CommandRunner(abc.ABC):
    """Run commands on the system and return relevant info."""

    @abc.abstractmethod
    def run(self, command: str, raise_on_command_fail: bool = True) -> CommandResult:
        """Run the command on the system.

        Args:
            command: The command to run.
            raise_on_command_fail: Whether to raise an error if the command is not successful.

        Raises:
            CommandFailedException when the command returns a nonzero exit code and raise_on_command_fail is True.
        """


class SubprocessCommandRunner(CommandRunner):
    """Use the subprocess module to run commands."""

    def run(self, command: str, raise_on_command_fail: bool = True) -> CommandResult:
        # Invoke the command on the system.
        subprocess_result = subprocess.run(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        # Create the result object.
        command_result = CommandResult(
            exit_code=subprocess_result.returncode,
            stdout=subprocess_result.stdout.decode(),
            stderr=subprocess_result.stderr.decode(),
        )
        # Check exit code and fail if appropriate.
        if subprocess_result.returncode != 0 and raise_on_command_fail:
            raise CommandRunnerException(command_result)
        # Return result.
        return command_result



from typing import Literal

from pagekey_semver.models import Tag
from pagekey_semver.file_replacer.base import FileReplacer, FileReplacerType
from pagekey_semver.util.command_runner import CommandRunner, SubprocessCommandRunner


class SedFileReplacer(FileReplacer):
    """Represents a file to replaced using `sed` on new release."""

    type: Literal[FileReplacerType.SED] = FileReplacerType.SED
    script: str

    def perform_replace(
        self, tag: Tag, runner: CommandRunner = SubprocessCommandRunner()
    ) -> str:
        """Run the sed program to replace a tag in a file.

        Replaces %M/%m/%p with tag's major/minor/patch,
        then calls sed with the provided `script`.

        Args:
            script: Sed script to run on the file.
        """
        # Check if the sed executable is available.
        result = runner.run("which sed", raise_on_command_fail=False)
        if result.exit_code != 0:
            raise EnvironmentError(
                "Sed executable not found on system - have you installed sed?"
            )
        # Replace placeholders in script using tag's major/minor/patch.
        script_escaped = self.script.replace("'", "\\'")
        script_replaced = (
            script_escaped.replace("%M", str(tag.major))
            .replace("%m", str(tag.minor))
            .replace("%p", str(tag.patch))
        )
        # Run sed.
        runner.run(f"sed -i '{script_replaced}' {self.name}")



import abc
import enum
from pydantic import BaseModel, field_serializer

from pagekey_semver.models import Tag


class FileReplacerType(str, enum.Enum):
    """File types supported for `file_replacers` in config."""

    JSON = "json"
    SED = "sed"
    TOML = "toml"
    YAML = "yaml"


class FileReplacer(BaseModel, abc.ABC):
    """File to be replaced on new release."""

    name: str
    type: FileReplacerType

    @field_serializer("type")
    def get_enum_value(self, v, info) -> str:
        """Turn the type field into a string when serializing."""
        return str(v.value)

    @abc.abstractmethod
    def perform_replace(self, tag: Tag) -> str:
        """Parse the `name` filepath and replace some part of it with the attributes of `tag`.

        Args:
            tag: The version tag to be used when replacing part of the file.
        """



"""Module for making changes to Git."""

import abc

from pagekey_semver.util.command_runner import CommandRunner, SubprocessCommandRunner


class GitEffector(abc.ABC):
    """Make changes to the current Git repo."""

    @abc.abstractmethod
    def set_config_item(self, key: str, value: str) -> None:
        """Set the Git config stored for the provided key.

        Args:
            key: Config key to set.
            value: Value to set.
        """

    @abc.abstractmethod
    def add_all(self) -> None:
        """Stage all current changes for commit."""

    @abc.abstractmethod
    def create_commit(self, message: str) -> None:
        """Make a new commit with the provided message.

        Args:
            message: The commit message.
        """

    @abc.abstractmethod
    def create_tag(self, name: str) -> None:
        """Create a tag with the provided name.

        Args:
            name: Name of tag to create.
        """

    @abc.abstractmethod
    def push(self, remote: str, ref: str) -> None:
        """Push a ref to a remote.

        Args:
            remote: Where to push.
            ref: Which commit, branch, or tag to push.
        """

    def checkout(self, ref: str) -> None:
        """Check out a ref.

        Args:
            ref: Which commit, branch, or tag to checkout.
        """

    def fetch_tags(self) -> None:
        """Fetch tags from all remotes."""


class CommandGitEffector(GitEffector):
    """Use the Git CLI to make changes to Git."""

    def __init__(self, command_runner: CommandRunner = SubprocessCommandRunner()):
        self._runner = command_runner

    def set_config_item(self, key: str, value: str) -> None:
        self._runner.run(f'git config {key} "{value}"')

    def add_all(self) -> None:
        self._runner.run("git add --all")

    def create_commit(self, message: str) -> None:
        self._runner.run(f'git commit -m "{message}"')

    def create_tag(self, name: str) -> None:
        self._runner.run(f"git tag {name}")

    def push(self, remote: str, ref: str) -> None:
        self._runner.run(f"git push {remote} {ref}")

    def checkout(self, ref: str) -> None:
        self._runner.run(f"git checkout {ref}")

    def fetch_tags(self) -> None:
        self._runner.run("git fetch --tags")



import abc
import os
from typing import Optional

import requests
from pydantic import BaseModel

from pagekey_semver.models import Tag


class CreateReleaseConfig(BaseModel):
    """."""

    project: str
    token_variable: str
    title_format: str
    body: str


class ReleaseCreator(abc.ABC):
    """Create a release on a Git hosting platform (GitHub, GitLab)."""

    def create_release(self, release_config: CreateReleaseConfig, tag: Tag):
        """."""


class GitHubIntegrationConfig(BaseModel):
    """Configuration for any GitHub-specific integrations."""

    create_release: Optional[CreateReleaseConfig] = None


class GitLabIntegrationConfig(BaseModel):
    """Configuration for any GitLab-specific integrations."""

    create_release: Optional[CreateReleaseConfig] = None


class GitHubReleaseCreator(ReleaseCreator):
    def create_release(self, release_config: CreateReleaseConfig, tag: Tag):
        token = os.getenv(release_config.token_variable, "")
        if len(token) < 1:
            print(
                f"WARNING: token {release_config.token_variable} is blank. Release creator may fail."
            )
        release_title = (
            release_config.title_format.replace("%M", str(tag.major))
            .replace("%m", str(tag.minor))
            .replace("%p", str(tag.patch))
        )
        print("POSTing GitHub release")
        request = requests.post(
            f"https://api.github.com/repos/{release_config.project}/releases",
            json={
                "tag_name": tag.name,
                "name": release_title,
                "body": release_config.body,
                "draft": False,
                "prerelease": False,
            },
            headers={
                "Authorization": f"token {token}",
                "Content-Type": "application/json",
            },
        )
        print(f"Status: {request.status_code}")
        print(f"Body: {request.content.decode()}")


class GitLabReleaseCreator(ReleaseCreator):
    def create_release(self, release_config: CreateReleaseConfig, tag: Tag):
        token = os.getenv("GITLAB_TOKEN")
        release_title = (
            release_config.title_format.replace("%M", str(tag.major))
            .replace("%m", str(tag.minor))
            .replace("%p", str(tag.patch))
        )
        print("POSTing GitLab release")
        request = requests.post(
            f"https://gitlab.com/api/v4/projects/{release_config.project}/releases",
            json={
                "tag_name": tag.name,
                "name": release_title,
                "description": release_config.body,
                "ref": tag.name,
            },
            headers={
                "PRIVATE-TOKEN": token,
                "Content-Type": "application/json",
            },
        )
        print(f"Status: {request.status_code}")
        print(f"Body: {request.content.decode()}")



"""Module related to config files."""

import os
from pathlib import Path
from typing import Annotated

from pydantic import BaseModel, Field
import yaml

from pagekey_semver.integrations.release_creator import (
    GitHubIntegrationConfig,
    GitLabIntegrationConfig,
)
from pagekey_semver.models import (
    GitConfig,
    Prefix,
)
from pagekey_semver.file_replacer.json import JsonFileReplacer
from pagekey_semver.file_replacer.sed import SedFileReplacer
from pagekey_semver.file_replacer.toml import TomlFileReplacer
from pagekey_semver.file_replacer.yaml import YamlFileReplacer
from pagekey_semver.util.env_to_dict import convert_env_to_dict
from pagekey_semver.util.update_dict import merge_dicts


class IntegrationsConfig(BaseModel):
    """Holds all configs for integrating with external services."""

    github: GitHubIntegrationConfig = GitHubIntegrationConfig()
    gitlab: GitLabIntegrationConfig = GitLabIntegrationConfig()


FileReplacersUnion = (
    JsonFileReplacer | SedFileReplacer | TomlFileReplacer | YamlFileReplacer
)


class SemverConfig(BaseModel):
    """Represents config options for entire application."""

    changelog_path: str
    changelog_writer: str
    format: str
    git: GitConfig
    prefixes: list[Prefix]
    file_replacers: list[Annotated[FileReplacersUnion, Field(discriminator="type")]]
    integrations: IntegrationsConfig = IntegrationsConfig()


DEFAULT_CONFIG = SemverConfig(
    changelog_path="CHANGELOG.md",
    changelog_writer="pagekey_semver.changelog_writer:DefaultChangelogWriter",
    format="v%M.%m.%p",
    git=GitConfig(
        name="PageKey Semver",
        email="semver@pagekey.io",
    ),
    prefixes=[
        Prefix(label="major", type="major"),
        Prefix(label="minor", type="minor"),
        Prefix(label="feat", type="minor"),
        Prefix(label="fix", type="patch"),
    ],
    file_replacers=[],
)
DEFAULT_CONFIG_DICT = DEFAULT_CONFIG.model_dump()


def load_config(config_path: Path) -> SemverConfig:
    """Load config from default, file, environment (in that order).

    Args:
        config_path: Path to config to parse.

    Returns:
        SemverConfig representing merged config from defaults, config file, and environment settings.
    """
    # Get the file config.
    if config_path.is_file():
        with open(config_path, "r") as file_handle:
            custom_config = yaml.safe_load(file_handle.read())
        config_without_env = merge_dicts(DEFAULT_CONFIG_DICT, custom_config)
    else:
        config_without_env = DEFAULT_CONFIG_DICT

    # Get the config defined by environment variables.
    env_config = convert_env_to_dict(os.environ)

    # Merge the file config and the environment config.
    # Environment takes precedence.
    final_config = merge_dicts(config_without_env, env_config)
    return SemverConfig(**final_config)



"""Module for CLI."""

import argparse
import os
import sys
from pathlib import Path

from pagekey_semver.changelog_writer import ChangelogWriter
from pagekey_semver.git.manager import GitManager
from pagekey_semver.integrations.release_creator import (
    GitHubReleaseCreator,
    GitLabReleaseCreator,
)
from pagekey_semver.release import SemverRelease
from pagekey_semver.models import Tag
from pagekey_semver.config import load_config


def cli_entrypoint(args=sys.argv[1:]):
    """Runs pagekey-semver when called from the command-line.

    Args:
        args: List of command-line args passed by user, or fake args for testing purposes.
    """
    parser = argparse.ArgumentParser(description="PageKey Semver")
    subparsers = parser.add_subparsers(title="Commands", dest="command")
    subparsers.add_parser(
        "plan", help="Compute which version would be created. (Dry-run)"
    )
    subparsers.add_parser("apply", help="Compute version, then commit, tag, and push.")
    parsed_args = parser.parse_args(args)

    if parsed_args.command == "plan":
        dry_run = True
    elif parsed_args.command == "apply":
        dry_run = False
    else:
        parser.print_usage()
        return

    config = load_config(Path(".semver"))

    # Init classes.
    manager = GitManager(config)
    release = SemverRelease(config)
    writer = ChangelogWriter.from_config(config)

    # Check out branch on GitLab
    gitlab_branch = os.getenv("CI_COMMIT_BRANCH", "")
    if len(gitlab_branch) > 0:
        print(f"Checking out CI_COMMIT_BRANCH={gitlab_branch}")
        # Yeah, yeah, bad form to use a private var this way. Too bad.
        manager._effector.checkout(gitlab_branch)
        manager._effector.fetch_tags()

    # Compute tags, commits
    tags = manager.get_git_tags()
    print(f"Found tags: {' '.join(tags)}")
    max_tag: Tag = release.get_biggest_tag(tags)
    hash = None
    if max_tag is not None:
        hash = max_tag.name
    commits = manager.get_commit_messages_since(hash)
    release_type = release.compute_release_type(commits)
    next_version: Tag = release.compute_next_version(release_type, tags)

    print("Next version:", next_version, flush=True)
    if max_tag == next_version:
        print("No new release (nothing to do).")
        if "GITHUB_OUTPUT" in os.environ:
            with open(os.environ["GITHUB_OUTPUT"], "w") as f:
                f.write("semver_release_occurred=false")
    else:
        # Apply tag if appropriate.
        if not dry_run:
            print(f"Applying version {next_version.name}.")
            # Write to changelog.
            writer.update_changelog(next_version, commits)
            # File Replacers
            print("Running file replacers.")
            for file_replacer in config.file_replacers:
                print(f"  {file_replacer.name}")
                file_replacer.perform_replace(next_version)
            # Apply tag, commit, push
            manager.apply_tag(tags, next_version)
            # Set an env var for subsequent steps in GitHub Actions.
            if "GITHUB_OUTPUT" in os.environ:
                with open(os.environ["GITHUB_OUTPUT"], "w") as f:
                    f.write("semver_release_occurred=true")
            # Create releases if enabled.
            if config.integrations.github.create_release is not None:
                GitHubReleaseCreator().create_release(
                    config.integrations.github.create_release, next_version
                )
            if config.integrations.gitlab.create_release is not None:
                GitLabReleaseCreator().create_release(
                    config.integrations.gitlab.create_release, next_version
                )
        else:
            print(f"Would apply version {next_version.name}.", flush=True)
            print("Dry run mode - not applying version.", flush=True)
            if "GITHUB_OUTPUT" in os.environ:
                with open(os.environ["GITHUB_OUTPUT"], "w") as f:
                    f.write("semver_release_occurred=false")


if __name__ == "__main__":
    cli_entrypoint()



import toml
from typing import Literal

from pagekey_semver.models import Tag
from pagekey_semver.file_replacer.base import FileReplacer, FileReplacerType
from pagekey_semver.util.update_dict import set_dict_value


class TomlFileReplacer(FileReplacer):
    """Represents TOML file to replaced on new release."""

    type: Literal[FileReplacerType.TOML] = FileReplacerType.TOML
    key: str
    format: str

    def perform_replace(self, tag: Tag) -> str:
        """Replace a key in a TOML file using the provided format.

        Args:
            tag: Version tag to replace key with.
        """
        # Read the file.
        with open(self.name, "r") as file_handle:
            contents = toml.load(file_handle)

        # Compute the value based on format and tag.
        new_version_str = (
            self.format.replace("%M", str(tag.major))
            .replace("%m", str(tag.minor))
            .replace("%p", str(tag.patch))
        )
        # Replace the key with the computed value.
        set_dict_value(contents, self.key, new_version_str)

        # Write the new file contents back to the same file.
        with open(self.name, "w") as file_handle:
            toml.dump(contents, file_handle)



"""Module for dynamically importing Python objects."""

import importlib
import os
import sys
from typing import Any


def dynamic_import(path: str) -> Any:
    """Import a Python object from a module import path.

    Note that this function automatically adds the current directory to `sys.path`,
    which allows `.py` files in the current directory to be imported.

    Args:
        path: A string representing a module import.
            A colon is used to separate the module from the object with it being imported.
            Example: "shutil:rmtree" will import the rmtree function from the shutil module and return it.

    Returns:
        The object represented by the "path" arg.

    Raises:
        ImportError, AttributeError if the import string is not valid.
    """
    # Include .py files in the current directory.
    sys.path.insert(0, os.getcwd())
    module_path, class_name = path.split(":")
    module = importlib.import_module(module_path)
    return getattr(module, class_name)



"""Module for querying Git."""

import abc
from typing import List, Optional

from pagekey_semver.models import Commit
from pagekey_semver.util.command_runner import (
    CommandRunner,
    CommandRunnerException,
    SubprocessCommandRunner,
)


class GitQuerier(abc.ABC):
    """Query the current Git repo for data."""

    @abc.abstractmethod
    def get_config_item(self, key: str) -> str:
        """Get the Git config stored for the provided key.

        Args:
            key: Config key to retrieve.

        Returns:
            The value of the config key.
        """

    @abc.abstractmethod
    def get_tag_names(self) -> List[str]:
        """Get a list of all Git tag names for the current repo.

        Returns:
            List of each tag name as a string.
        """

    @abc.abstractmethod
    def get_commits(self, since_ref: Optional[str] = None) -> List[str]:
        """Return commits in this repo since provided ref, or all commits.

        Args:
            since_ref: Reference to a commit to pull messages since, or None.

        Returns:
            Commits since `commit_hash` if provided.
            All commits if `commit_hash` is None.
        """


class CommandGitQuerier(GitQuerier):
    """Use the Git CLI to query Git."""

    def __init__(self, command_runner: CommandRunner = SubprocessCommandRunner()):
        self._runner = command_runner

    def get_config_item(self, key: str):
        try:
            result = self._runner.run(f"git config {key}")
            return result.stdout.strip()
        except CommandRunnerException:
            print(f"Failed to retrieve git config key {key}")
            return ""

    def get_tag_names(self) -> List[str]:
        result = self._runner.run("git tag")
        # Filter out empty strings and return.
        return [tag for tag in result.stdout.split() if len(tag) > 0]

    def get_commits(self, since_ref: Optional[str] = None) -> List[str]:
        # Create command.
        command = 'git log --pretty="format:%H %s"'
        if since_ref is not None and len(since_ref) > 0:
            command += f" {since_ref}..HEAD"
        # Run the command.
        command_result = self._runner.run(command)
        # Parse the commit messages.
        commits = []
        for commit_log in command_result.stdout.split("\n"):
            # Ignore blank lines.
            if len(commit_log) == 0:
                continue
            # Separate the hash from the commit message.
            fields = commit_log.split()
            commits.append(
                Commit(
                    hash=fields[0],
                    message=" ".join(fields[1:]),
                )
            )
        return commits



"""Module for functions that modify dicts."""

from typing import Any


def get_dict_value(d: dict, path: str, sep: str = ".") -> Any:
    """
    Retrieve a value from a nested dictionary using a dot-separated (or other separator) string path.

    Args:
        d: The dictionary to retrieve the value from.
        path: The path string specifying the keys, separated by the given separator.
        sep: The separator used in the path string. Defaults to ".".

    Returns:
        The value from the dictionary at the specified path.

    Raises:
        KeyError: If any of the keys in the path do not exist in the dictionary.
    """
    keys = path.split(sep)
    for key in keys:
        d = d[key]
    return d


def set_dict_value(d: dict, path: str, value: Any, sep: str = ".") -> None:
    """
    Set a value in a nested dictionary using a dot-separated (or other separator) string path.

    Args:
        d (dict): The dictionary to set the value in.
        path (str): The path string specifying the keys, separated by the given separator.
        value: The value to set at the specified path.
        sep (str): The separator used in the path string. Defaults to ".".

    Returns:
        None. The function modifies the dictionary in place.

    Raises:
        TypeError: If the intermediate values in the path are not dictionaries.
    """
    keys = path.split(sep)
    for key in keys[:-1]:
        d = d.setdefault(key, {})
    d[keys[-1]] = value


def merge_dicts(d1: dict, d2: dict) -> dict:
    """
    Merges two dictionaries recursively. If both dictionaries contain a key
    with a dictionary as its value, the function will recursively merge the
    sub-dictionaries. If a key exists in both dictionaries but the corresponding
    values are not dictionaries, the value from the second dictionary (d2)
    will overwrite the value from the first dictionary (d1).

    Args:
        d1 (dict): The first dictionary to merge.
        d2 (dict): The second dictionary to merge.

    Returns:
        dict: A new dictionary that is the result of merging d1 and d2.
    """
    # Start with a copy of the first dictionary
    merged = d1.copy()

    for key, value in d2.items():
        if key in merged and isinstance(merged[key], dict) and isinstance(value, dict):
            # If both d1 and d2 have the same key and the value is a dict, merge them recursively
            merged[key] = merge_dicts(merged[key], value)
        else:
            # Otherwise, just set the value from d2
            merged[key] = value

    return merged
